/*
 * Virtual Drum Machine - Rhythm Compiler runtime library
 * This project is in the public domain.
 *---
 * The main file
 */

#include "card.h"
#include "mixer.h"
#include "sound.h"
#include "wav.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* one global, for simplicity reasons */
SOUND sound;

/* another global, shame shame shame */
double vol = 1.;	/* the global volume, default value 1 */

/* this function must be defined by the user */
extern void main_rhythm(int, char **);

/* this functions will be generated by rc */
extern void init_samples(void);
extern void close_samples(void);

void usage(FILE *f, char *p)
{
  fprintf(f, "%s: this program was produced by rc (version %s), the Rhythm Compiler\n", p, VERSION);
  fprintf(f, "options:\n");
  fprintf(f, "    -h | --help : to print this help screen\n");
  fprintf(f, "    -v | --version : to print the version of rc\n");
  fprintf(f, "    -w <file>   : to dump to a wav file (filename can't start by -)\n");
  fprintf(f, "    -r <file>   : to dump to a raw file (filename can't start by -)\n");
  fprintf(f, "    -r2 <file1> <file2> : to dump to two raw file\n");
  fprintf(f, "                          (this option sets the stereo mode,\n");
  fprintf(f, "                           you can't use -s with it)\n");
  fprintf(f, "                          (filename can't start by -)\n");
  fprintf(f, "    -d <device> : to choose the device (\"%s\" by default)\n", CARD_DEFAULT_DEV);
  fprintf(f, "    -f <freq>   : to choose the output frequency (%d by default))\n", CARD_DEFAULT_FREQ);
  fprintf(f, "                  (must be between 5000 and 44100)\n");
  fprintf(f, "    -s          : to choose stereo mode (mono by default)\n");
  fprintf(f, "                  (don't use it with -r2)\n");
  fprintf(f, "-w, -r, -r2, -d are exclusive options !\n");
}

int main(int n, char **v)
{
  char **nv = malloc(n * sizeof(char *));
  int nn = 0;
  int i;
  int set_output=0;
  int set_freq=0;
  int set_stereo=0;

  if (!nv) {
    perror("malloc");
    return -1;
  }

  nv[0] = v[0];
  nn++;

  /* default parameters */
  /* card */
  sound.card.mode = CARD_MONO;
  sound.card.device = CARD_DEFAULT_DEV;
  sound.card.frequency = CARD_DEFAULT_FREQ;

  /* mixer */
  sound.mixer.mode = MIXER_MONO;
  sound.mixer.tempo = 60;
  sound.mixer.frequency = CARD_DEFAULT_FREQ;

  /* wav */
  sound.wav.mode = WAV_DO_NOTHING;
  sound.wav.frequency = CARD_DEFAULT_FREQ;
  sound.wav.channels = 1;
  sound.wav.bit_resolution = 16;

  /* handle params */ 
  for (i=1; i<n; i++) {

    if (!strcmp(v[i], "-v") || !strcmp(v[i], "--version")) {
      fprintf(stdout, "produced by rc version %s\n", VERSION);
      goto end;
    }

    if (!strcmp(v[i], "-h") || !strcmp(v[i], "--help")) {
      usage(stdout, v[0]);
      goto end;
    }

    if (!strcmp(v[i], "-w")) {
      if (set_output) goto error_param;
      set_output = 1;
      i++;
      if (i == n) goto error_param;
      sound.card.mode = CARD_NO_SOUND;
      sound.wav.mode = WAV_SAVE_WAV;
      /* we avoid filenames starting by - */
      if (v[i][0] == '-') goto error_param;
      sound.wav.filename = v[i];
      continue;
    }

    if (!strcmp(v[i], "-r")) {
      if (set_output) goto error_param;
      set_output = 1;
      i++;
      if (i == n) goto error_param;
      sound.card.mode = CARD_NO_SOUND;
      sound.wav.mode = WAV_SAVE_RAW;
      /* we avoid filenames starting by - */
      if (v[i][0] == '-') goto error_param;
      sound.wav.filename = v[i];
      continue;
    }

    if (!strcmp(v[i], "-r2")) {
      if (set_output || set_stereo) goto error_param;
      set_output = 1;
      set_stereo = 1;
      i++;
      if (i == n || i == n-1) goto error_param;
      sound.card.mode = CARD_NO_SOUND;
      sound.wav.mode = WAV_SAVE_RAW2;
      sound.mixer.mode = MIXER_SEPARATED;
      /* we avoid filenames starting by - */
      if (v[i][0] == '-') goto error_param;
      if (v[i+1][0] == '-') goto error_param;
      sound.wav.filename = v[i];
      sound.wav.filename2 = v[i+1];
      i++;
      continue;
    }

    if (!strcmp(v[i], "-d")) {
      if (set_output) goto error_param;
      set_output = 1;
      i++;
      if (i == n) goto error_param;
      sound.card.device = v[i];
      continue;
    }

    if (!strcmp(v[i], "-f")) {
      unsigned int f;
      char *e;
      if (set_freq) goto error_param;
      set_freq = 1;
      i++;
      if (i == n) goto error_param;
      if (strlen(v[i]) == 0) goto error_param;
      f = strtoul(v[i], &e, 0);
      if (*e) goto error_param;
      if (f < 5000 || f > 44100) goto error_param;
      sound.card.frequency = f;
      sound.mixer.frequency = f;
      sound.wav.frequency = f;
      continue;
    }

    if (!strcmp(v[i], "-s")) {
      if (set_stereo) goto error_param;
      set_stereo = 1;
      sound.mixer.mode = MIXER_STEREO;
      sound.wav.channels = 2;
      continue;
    }

    /* no recognized option, pass it to main_rhythm */
    nv[nn] = v[i];
    nn++;
  }

  if (sound.mixer.mode == MIXER_STEREO && sound.card.mode == CARD_MONO)
    sound.card.mode = CARD_STEREO;

  /* ok, let's do it */
  if (init_sound(&sound)==-1)
    goto end2;

  init_samples();

  main_rhythm(nn, nv);
  finalize_sound(&sound);

  close_samples();

end2:
  close_sound(&sound);

end:
  free(nv);

  return 0;

error_param:
  free(nv);

  usage(stderr, v[0]);
  return 1;
}
