/*
 * Virtual Drum Machine - Rhythm Compiler runtime library
 * This project is in the public domain.
 *---
 * The mixer stuff
 */

#include "mixer.h"

#include <stdio.h>
#include <stdlib.h>

#include "sample.h"

int init_mixer(MIXER *m)
{
  m->list = 0;
  m->free_list = 0;
  m->buffer = 0;
  m->buffer_out = 0;
  m->buffer_size = 0;

  return 0;
}

void close_mixer(MIXER *m)
{
  CHANNEL_LIST *l;

  /* free the buffers */
  if (m->buffer_size) {
    free(m->buffer);
    free(m->buffer_out);
  }

  /* free the lists */
  l = m->list;
  while(l) {
    CHANNEL_LIST *d=l->next;
    free(l);
    l = d;
  }

  l = m->free_list;
  while(l) {
    CHANNEL_LIST *d=l->next;
    free(l);
    l = d;
  }
}

/* size is in bytes */
/* return 0 if end of buffer, 1 if some more data to handle after this call */
static int mix_mono(CHANNEL *c, register double *buf, int size)
{
  register short *data=&c->data[c->proceeded_size/2];
  register double f;
  register double volume = c->vol;

  if (size > c->size - c->proceeded_size)
    size = c->size - c->proceeded_size;

  if (size == 0)
    return 0;

  c->proceeded_size += size;

  size /= 2;

  for (; size; size--) {
    f = (*buf) + (double)(*data) * volume;

    *buf = f;

    buf++;
    data++;
  }

  if (c->proceeded_size >= c->size)
    return 0;

  return 1;
}

/* size is in bytes */
/* return 0 if end of buffer, 1 if some more data to handle after this call */
static int mix_stereo(CHANNEL *c, register double *buf, int size)
{
  register short *data=&c->data[c->proceeded_size/2];
  register double f;
  register double volume = c->vol;
  register double lv, rv;

  size /= 2;  /* we will only proceed size/2 input bytes */

  if (size > c->size - c->proceeded_size)
    size = c->size - c->proceeded_size;

  if (size == 0)
    return 0;

  c->proceeded_size += size;

  size /= 2;

  /* setup left volume */
  if (c->pan <= 0)
    lv = 1;
  else
    lv = 1 - c->pan;

  /* right volume */
  if (c->pan >= 0)
    rv = 1;
  else
    rv = 1 + c->pan;

  lv *= volume;
  rv *= volume;

  for (; size; size--) {
    f = (double)(*data);

    *buf += f * lv;
    *(buf+1) += f * rv;

    buf++; buf++;
    data++;
  }

  if (c->proceeded_size >= c->size)
    return 0;

  return 1;
}

/* size is in bytes */
/* the difference with mix_stereo is both buffers are separated,
 * ie the mixing buffer will be : [..left data.. | ..righ data..]
 */
/* return 0 if end of buffer, 1 if some more data to handle after this call */
static int mix_separated(CHANNEL *c, register double *buf, int size)
{
  int saved_size = size;
  register short *data=&c->data[c->proceeded_size/2];
  register double f;
  register double volume = c->vol;
  register double lv, rv;
  register double *buf2;

  /* useless check ? */
  if (size & 3) {
    fprintf(stderr, "impossible error, contact sed@free.fr\n");
    abort();
  }

  size /= 2;  /* we will only proceed size/2 input bytes */

  if (size > c->size - c->proceeded_size)
    size = c->size - c->proceeded_size;

  if (size == 0)
    return 0;

  c->proceeded_size += size;

  size /= 2;

  /* setup left volume */
  if (c->pan <= 0)
    lv = 1;
  else
    lv = 1 - c->pan;

  /* right volume */
  if (c->pan >= 0)
    rv = 1;
  else
    rv = 1 + c->pan;

  lv *= volume;
  rv *= volume;

  /* will one believe I spent 2 days on this ? I did buf2 = buf+size; but size was wrong when end of sample came, damned... 2 days... */
  buf2 = buf + saved_size/4;

  for (; size; size--) {
    f = (double)(*data);
    *buf += f * lv;
    *buf2 += f * rv;

    buf++;
    buf2++;
    data++;
  }

  if (c->proceeded_size >= c->size)
    return 0;

  return 1;
}

/* size is in bytes */
static void post_mix(MIXER *m, int size)
{
  register short *b = m->buffer_out;
  register double *z = m->buffer;
  register int i;
  register int nb_samples = size/2;
  register double f;

  for (i=0; i<nb_samples; i++, z++, b++) {
    f = *z;
    if (f > 32767)
      f = 32767;
    if (f < -32768)
      f = -32768;
    *b = f;
  }
}

/* size is in bytes */
void mix_some(MIXER *m, int size)
{
  int (*mix_me)(CHANNEL *, double *, int);
  CHANNEL_LIST *cur, *pre, *dum;
  unsigned long nb_samples = size/2;

  /* useless check ? */
  if ((m->mode == MIXER_STEREO || m->mode == MIXER_SEPARATED) && (nb_samples&1)) {
    fprintf(stderr, "error, mixer is in stereo, but samples nb is not = 0 mod 2\n");
    fprintf(stderr, "contact sed@free.fr NOW !\n");
    exit(1);
  }

  if (m->mode == MIXER_MONO)
    mix_me = mix_mono;
  else if (m->mode == MIXER_STEREO)
    mix_me = mix_stereo;
  else if (m->mode == MIXER_SEPARATED)
    mix_me = mix_separated;
  else {
    fprintf(stderr, "error, bad mixing mode %d\n", m->mode);
    exit(1);
  }

  if (m->buffer_size >= nb_samples)
    goto buffer_ok;

  m->buffer = realloc(m->buffer, nb_samples * sizeof(double));
  if (!m->buffer) {
    perror("mix_some");
    exit(1);
  }

  m->buffer_out = realloc(m->buffer_out, nb_samples * sizeof(short));
  if (!m->buffer_out) {
    perror("mix_some");
    exit(1);
  }

  m->buffer_size = nb_samples;

buffer_ok:
  /* we suppose that 0 as a double is all bits at 0, maybe false ? (should check the specs)*/
  memset(m->buffer, 0, nb_samples * sizeof(double));

  pre = 0;
  cur = m->list;

  while(cur) {
    if (mix_me(&cur->channel, m->buffer, size)==0) {
      /* channel free, let's put it into free list */
      if (pre)
	pre->next = cur->next;
      else
	m->list = cur->next;
      dum=cur->next;
      cur->next = m->free_list;
      m->free_list = cur;
      cur = dum;
    } else {
      pre = cur;
      cur = cur->next;
    }
  }

  post_mix(m, size);
}

void mixer_add_channel(MIXER *m, SAMPLE *s, double volume, double panning)
{
  CHANNEL_LIST *cur;

  if (m->free_list)
    goto free_list_ok;

  m->free_list = malloc(sizeof(CHANNEL_LIST));
  if (!m->free_list) {
    perror("mixer_add_channel");
    exit(1);
  }

  m->free_list->next = 0;

free_list_ok:
  /* take out head of free list */
  cur = m->free_list;
  m->free_list = m->free_list->next;

  /* put in head of played list */
  cur->next = m->list;
  m->list = cur;

  /* init inside stuff */
  cur->channel.data = s->buffer;
  cur->channel.size = s->size;
  cur->channel.proceeded_size = 0;
  cur->channel.vol = volume * vol;	/* vol is global */
  cur->channel.pan = panning;
}
