/*
 * Virtual Drum Machine - Rhythm Compiler program
 * This project is in the public domain.
 *---
 * The compiler definition file
 */

#include "compiler.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE	/* we use getline */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <ctype.h>

#include "rhythm.h"
#include "environment.h"
#include "error.h"

char *header="/*\n"
" * Virtual Drum Machine - Rhythm Compiler program\n"
" * This project is in the public domain.\n"
" *---\n"
" * This C file was generated by the Rhythm Compiler from the file \"%s\".\n"
" */\n"
"\n"
"#include \"defines.h\"\n"
"\n";

/* return 0 if ok, <0 if error */
int init_compiler(COMPILER *compiler, int n, char **v, ENVIRONMENT *env)
{
  int o = 0, c = 0;
  int curopt = 1;
  compiler->input_file = 0;
  compiler->output_file = "a.out";
  compiler->c_only = 0;

  /* let's parse the options */
  while (curopt < n) {
    /* it can be == 0, because of environment handling some of them and setting to 0 if it parsed it */
    if (!v[curopt]) {
      curopt++;
      continue;
    }

    if (v[curopt][0] != '-') {
      /* it's not an option, so must be the input file */
      if (compiler->input_file)
	return -INPUT_FILE_ALREADY_DEFINED;
      compiler->input_file=v[curopt];
      curopt++;
      continue;
    }
    /* so it's an option, let's parse it */
    switch (v[curopt][1]) {
      case 'o' :
	if (v[curopt][2])
	  return -BAD_OPTION_NAME;
	if (o)
	  return -OUTPUT_FILE_ALREADY_DEFINED;
	if (curopt+1==n)
	  return -FILE_NAME_EXPECTED;
	curopt++;
	o=1;
	compiler->output_file=v[curopt];
	break;
      case 'c' :
	if (v[curopt][2])
	  return -BAD_OPTION_NAME;
	if (c)
	  return -C_ONLY_ALREADY_DEFINED;
	compiler->c_only = 1;
	break;
      default :
	return -BAD_OPTION_NAME;
    }
    curopt++;
  }

  if (!compiler->input_file)
    return -INPUT_FILE_REQUIRED;

  return 0;
}


int run_gcc(char *file, COMPILER *c, ENVIRONMENT *env)
{
  int status;

  /* let's fork */
  int f=fork();

  if (f==-1)
    return -FORK_ERROR;

  if (f != 0)
    goto wait_gcc;

  execlp("gcc", "gcc", "-x", "c", "-I", env->inc_dir, "-L", env->lib_dir, "-o", c->output_file, file, "-lvdm", (char *)0);
  perror("execlp");

  exit(1);

wait_gcc:
  /* maybe not very accurate, huh ? */
  wait(&status);

  if (!WIFEXITED(status))
    return -GCC_ERROR;

  if (WEXITSTATUS(status))
    return -GCC_ERROR;

  return 0;
}

/* return 1 if the given line is a rhythm line, 0 if not */
int is_rhythm_line(char *line, int size)
{
  while (size && isspace(*line)) { size--; line++; }

  if (!size) return 0;

  if (isnote(*line)) return 1;

  return 0;
}

/* error with fprintf (if disk full for instance) very badly handled */
/* return 0 if ok, <0 if error */
int compile(COMPILER *compiler, RHYTHM *rhythm, ENVIRONMENT *env)
{
  char tmpname[] = "/tmp/rc.i.XXXXXX";
  FILE *f;
  char *fname;
  FILE *input;
  char *line=0;
  size_t n=0;
  int i;
  ssize_t size;
  int current_line=0;
  int ret;

  if (compiler->c_only) {
    fname=compiler->output_file;
    f=fopen(fname, "wb");
  } else {
    int fd=mkstemp(tmpname);
    if (fd == -1)
      return -IMPOSSIBLE_CREATE_TEMP_NAME;
    fname = tmpname;
    f=fdopen(fd, "wb");
  }
  if (!f)
    return -CANT_CREATE_FILE;

  input=fopen(compiler->input_file, "r");
  if (!input) {
    fclose(f);
    unlink(fname);
    return -FILE_NOT_FOUND;
  }

  /* let's write the header of the C file */
  fprintf(f, "# 1 \"internal rc data file\"\n");
  fprintf(f, header, compiler->input_file);

  /* now, let's create our samples */
  /* first define them global - sbeyer */
  for (i=0; i<env->nb_elem; i++)
    fprintf(f, "SAMPLE %s;\n", env->vars[i].name);
  fprintf(f, "\n");

  /* ... sbeyer - ok, it's all bad style but who cares ;) */
  /* no it's not :-) - Sed addition */
  /* and now initialize them */
  fprintf(f, "/* initializing all our (global) samples */\n");
  fprintf(f, "void init_samples(void)\n{\n");
   for (i=0; i<env->nb_elem; i++)
    fprintf(f, "\tinit_sample(&%s, \"%s\");\n", env->vars[i].name, (char *)env->vars[i].value);
  fprintf(f, "}\n\n");

  /* and close them */
  fprintf(f, "/* free all our (global) samples */\n");
  fprintf(f, "void close_samples(void)\n{\n");
   for (i=0; i<env->nb_elem; i++)
    fprintf(f, "\tclose_sample(&%s);\n", env->vars[i].name);
  fprintf(f, "}\n\n");

  fprintf(f, "# 1 \"%s\"\n", compiler->input_file);

  /* now, we parse the input file, and if the line is a rhythm line, we put some orders, otherwise we put it in the
   * resulting file, the last note is not a real one, we can forget it
   */
  for (i=0; i < rhythm->nb_elem-1;) {
    int j;

    if ((size=getline(&line, &n, input))==-1)
      return -GETLINE_ERROR;

    current_line++;

    //if (current_line != rhythm->notes[i].line_in_file) {
    if (n == 0 || !is_rhythm_line(line, n-1)) {
      fprintf(f, "%s", line);
      continue;
    }

    while (current_line == rhythm->notes[i].line_in_file) {
      for (j=0; j < rhythm->notes[i].nb_elem; j++) {
	fprintf(f, "# %d \"%s\"\n", current_line, compiler->input_file);
	fprintf(f, "\tput_note(&sound, &%s, %s.vol, %s.pan);\n",
			rhythm->notes[i].instruments[j].name,
			rhythm->notes[i].instruments[j].name,
			rhythm->notes[i].instruments[j].name); /* sbeyer & Sed */
      }

      fprintf(f, "# %d \"%s\"\n", current_line, compiler->input_file);
      fprintf(f, "\tplay_some_sound(&sound, %d);\n", rhythm->notes[i].duration);

      i++;
    }
  }
  /* ok, it's bad, I should have my previous for on the lines of the file, not on the note,
   * but well, I am too lazy right to change it, so here is a final loop (we need the last lines)
   */
  fprintf(f, "# %d \"%s\"\n", ++current_line, compiler->input_file);
  while ((size=getline(&line, &n, input))!=-1)
    fprintf(f, "%s", line);

  if (fwrite("\n", 1, 1, f) != 1) {
    perror(fname);
    fclose(f);
    fclose(input);
    unlink(fname);
    return -ERROR_WRITING_FILE;
  }

  fclose(f);
  fclose(input);

  if (compiler->c_only) {
    fprintf(stdout, "Kept temporary C file in %s\n", fname); /* sbeyer */
    fprintf(stdout, "To compile, you can type:\n");
    fprintf(stdout, "\tgcc -x c -I %s -L %s -o <ExecutableOutput> %s -lvdm\n", env->inc_dir, env->lib_dir, fname);
    fprintf(stdout, "or remove it.");
    return 0;
  }

  /* let's call gcc */
  ret=run_gcc(fname, compiler, env);

  unlink(fname);

  if (ret != 0)
    return ret;

  return 0;
}
