/*
 * Virtual Drum Machine - Rhythm Compiler program
 * This project is in the public domain.
 *---
 * The environment definition file
 */

#include "environment.h"

#ifndef _GNU_SOURCE
#define _GNU_SOURCE		/* for getline */
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <unistd.h>
 
#include "error.h"
#include "memory.h"

char *copy_string(char *s, int size)
{
  char *ret = getmem(size+1);

  memcpy(ret, s, size);

  ret[size]=0;

  return ret;
}

/* return 0 if ok, -1 if not */
int env_add_sound_file(char *buf, ENVIRONMENT *env)
{
  struct stat stbuf;
  int size1=0, size2=0;
  char *buf1, *buf2;

  if (!isspace(*buf))
    return -1;

  while (*buf && isspace(*buf))
    buf++;

  if (!(*buf))
    return -1;

  buf1 = buf;

  while (*buf && !isspace(*buf))
    buf++, size1++;

  if (!(*buf))
    return -1;
  while (*buf && isspace(*buf))
    buf++;
  if (!(*buf))
    return -1;

  buf2 = buf;

  while (*buf && !isspace(*buf))
    buf++, size2++;

  /* so (buf1, size1) is the "nickname" of the file (buf2, size2), let's add it to our list */
  if (env->size == env->nb_elem)
    env->vars = regetmem(env->vars, env->size + sizeof(ENVVAR) * 32);
  env->size += 32;
  env->vars[env->nb_elem].name = copy_string(buf1, size1);
  env->vars[env->nb_elem].value = copy_string(buf2, size2);
  if (stat(env->vars[env->nb_elem].value, &stbuf)==-1)
    fprintf(stderr, "Configuration warning: the file %s does not exist\n", (char *)env->vars[env->nb_elem].value);
  env->nb_elem++;

  return 0;
}

/* return 0 if ok, -1 if not */
int env_add_include_dir(char *buf, ENVIRONMENT *env)
{
  struct stat stbuf;
  char *buf1;
  int size1 = 0;

  if (!isspace(*buf))
    return -1;

  while (*buf && isspace(*buf))
    buf++;

  if (!(*buf))
    return -1;

  buf1 = buf;

  while (*buf && !isspace(*buf))
    buf++, size1++;

  env->inc_dir = copy_string(buf1, size1);
  if (stat(env->inc_dir, &stbuf)==-1)
    fprintf(stderr, "Configuration warning: the directory %s does not exist\n", env->inc_dir);

  return 0;
}

/* return 0 if ok, -1 if not */
int env_add_lib_dir(char *buf, ENVIRONMENT *env)
{
  struct stat stbuf;
  char *buf1;
  int size1 = 0;

  if (!isspace(*buf))
    return -1;

  while (*buf && isspace(*buf))
    buf++;

  if (!(*buf))
    return -1;

  buf1 = buf;

  while (*buf && !isspace(*buf))
    buf++, size1++;

  env->lib_dir = copy_string(buf1, size1);
  if (stat(env->lib_dir, &stbuf)==-1)
    fprintf(stderr, "Configuration warning: the directory %s does not exist\n", env->lib_dir);

  return 0;
}

/* return 0 if ok, 1 if --help sollicited, 2 if --version wanted, <0 if error */
int init_environment(ENVIRONMENT *env, int argc, char **argv)
{
  FILE *conf;
  char *hn;
  char *nname;
  char *line=0;
  int i;
  int current_line = 0;

  memset(env, 0, sizeof(ENVIRONMENT));

  /* first of all, is there --help ? */
  for (i=1; i<argc; i++)
    if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help"))
      return 1;

  /* or --version ? */
  for (i=1; i<argc; i++)
    if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "--version"))
      return 2;

  /* no, let's open the config file (it must exist !) */
  /* it can be given as a command line argument (--config-file or -cf) */
  for (i=1; i<argc; i++) {
    if (strcmp(argv[i], "--config-file") && strcmp(argv[i], "-cf"))
      continue;

    i++;
    if (i == argc)
      return -NO_CONFIG_FILE_NAME;

    if ((conf=fopen(argv[i], "r"))) {
      /* is it legal (ansi) to modify argv ?? */
      fprintf(stderr, "using config file %s\n", argv[i]);
      argv[i-1]=0;
      argv[i]=0;
      goto config_file_found;
    }

    return -BAD_CONFIG_FILE_NAME;
  }

  /* well, no file given in command line, maybe it is in the current dir */
  if ((conf=fopen("rc.conf", "r"))) {
    fprintf(stderr, "using config file %s\n", "rc.conf");
    goto config_file_found;
  }

  /* then maybe $HOME/rc.conf */
  hn = getenv("HOME");
  if (!hn)
    goto not_in_home;

  nname = getmem(strlen(hn)+strlen("/rc.conf")+1);
  strcpy(nname, hn);
  strcat(nname, "/rc.conf");

  if ((conf=fopen(nname, "r"))) {
    fprintf(stderr, "using config file %s\n", nname);
    free(nname);
    goto config_file_found;
  }

not_in_home:
  /* well, so it must be in INSTALL_DIR dir */
  nname = getmem(strlen(INSTALL_DIR)+strlen("/rc.conf")+1);
  strcpy(nname, INSTALL_DIR);
  strcat(nname, "/rc.conf");

  if ((conf=fopen(nname, "r"))) {
    fprintf(stderr, "using config file %s\n", nname);
    free(nname);
    goto config_file_found;
  }

  /* nowhere, I can't proceed, installation badly done, let's quit */
  return -CONFIG_FILE_NOT_FOUND;

config_file_found:
  /* let's read it, line by line */
  /* we won't be too fascistic, if the beginning of a line matches nothing, we'll pass it through, no problem */
  /* maybe just a little warning, in case the user typed something wrong by mistake */
  /* commented lines begin with # (we'll be silent with those lines) */
  while (!feof(conf)) {
    ssize_t size;
    char *buf;
    size_t unused=0;

    current_line++;

    if ((size=getline(&line, &unused, conf))==-1 && !feof(conf))
      return -GETLINE_ERROR;
    if (size==-1)
      break;

    buf=line;
    while (*buf && isspace(*buf))
      buf++;
    if (!(*buf))
      continue;
    if (*buf== '#')
      continue;
    if (!strncmp(buf, "sample", 6)) {
      if (env_add_sound_file(buf+6, env)==-1)
	fprintf(stderr, "Warning, wrong line (%d) in configuration file :\n%s", current_line, line);
      continue;
    }
    if (!strncmp(buf, "INCLUDE_DIR", 11)) {
      if (env_add_include_dir(buf+11, env)==-1)
	fprintf(stderr, "Warning, wrong line (%d) in configuration file :\n%s", current_line, line);
      continue;
    }
    if (!strncmp(buf, "LIBRARY_DIR", 11)) {
      if (env_add_lib_dir(buf+11, env)==-1)
	fprintf(stderr, "Warning, wrong line (%d) in configuration file :\n%s", current_line, line);
      continue;
    }
    /* nothing appropriate found, let's warn the user */
    fprintf(stderr, "Warning, in the configuration file, line %d means nothing to me :\n%s", current_line, line);
  } /* while (!feof(conf)) */

  if (!env->inc_dir)
    return -ENV_INC_DIR_NOT_SET;

  if (!env->lib_dir)
    return -ENV_LIB_DIR_NOT_SET;

  return 0;
}
